/* Chromosome.cpp
   implemenation of Chromosome.hpp
   @author	Hyun Kim
   @date	11/20/99
*/

#include "Chromosome.hpp"
#include "random.hpp"
#include "Parameter.hpp"
#include "utilities.hpp"
#include <math.h>
#include <assert.h>

/* void ctor
*/
Chromosome::Chromosome()
{
	genes = new double[ Parameter::genes ];
	assert( genes != 0 );
	setLength( Parameter::genes );
}

/* copy ctor
*/
Chromosome::Chromosome( Chromosome& chrom )
{
	genes = new double[ Parameter::genes ];
	assert( genes != 0 );
	for ( int i = 0; i < Parameter::genes; i++ )
	{
		genes[ i ] = chrom[ i ];
	}
	fitness = chrom.getFitness();
	length = chrom.getLength();
}

/* dtor
*/
Chromosome::~Chromosome()
{
	delete[] genes;
}

/* return the point of the i-th gene in this chromosome
   @param i i-th gene
   @return double
*/
double& Chromosome::operator[]( int i )
{
	assert( i >= 0 && i < length );
	return genes[ i ];
}

/* assign the chromosome to another chromosome
   @param chrom chromosome for copying
   @return Chromosome
*/
Chromosome&	Chromosome::operator=( Chromosome& chrom )
{
	if( this == &chrom ) 
	{
		return *this;
	}

	delete[] genes;
	genes = new double[ Parameter::genes ];
	assert( genes != 0 );
	for ( int i = 0; i < Parameter::genes; i++ )
	{
		genes[ i ] = chrom[ i ];
	}
	fitness = chrom.getFitness();
	length = chrom.getLength();
	return *this;
}

/* It randomly distributes the value of the gene in the chromosome
   It meets the first constraint, i.e. sigma(Xi) = 1	where i = 1 to N
   To ensure the sum of the N weights equals 1 and all N weights are positive numbers 
   between 0 to 1, the following procedure is followed: 
   1. Generate N random numbers. 
   2. Assign the i-th random number divided by the sum of the N random numbers as the 
   weight for i-th stock.
*/
void Chromosome::random()
{
	double gene, total = 0;
	int i;

	for ( i = 0; i < getLength(); i++ )
	{	
		gene = random01();
		genes[ i ] = gene;
		total += gene;
	}
	for ( i = 0; i < getLength(); i++ )
	{	
		genes[ i ] /= total;
	}
}

/* compute the fitenss of the chromosome
*/
void Chromosome::evaluate()
{
	double myFitness = objectiveFunction();
	setFitness( myFitness );
}

/* return the point of any gene in this chromosome
   @return int
*/
int Chromosome::getRandomPoint()
{
	int point;
	point = rnd( 0, getLength() - 1 );
	return point;
}

/* exchange the value of any two genes in the chromosome
*/
void Chromosome::transpose()
{
	if ( random01() > Parameter::mutationRate )
	{
		return;
	}

	// transpose randomly chosen two genes within the mutation rate
	int p1 = getRandomPoint();
	int p2 = getRandomPoint();

	if ( p1 != p2 )
	{
		swap( genes[ p1 ], genes[ p2 ] );
	}
}

/* change the value of any gene. If the original value is 0.23, then mutated value is 0.77.
*/
void Chromosome::mutate()
{
	if ( random01() > Parameter::mutationRate )
	{
		return;
	}

	// mutate randomly chosen gene within the mutation rate
	int p = getRandomPoint();
	genes[ p ] = 1 - genes[ p ];
	//adjust( Parameter::genes );
}

/* adjust the value of each gene to meet the constraint of the problem
   In this problem, 
   constraint 1: all values of each gene should be one.
   constraint 2: the value of the gene should be positive.
   @param n number of genes which should be adjusted
*/
void Chromosome::adjust( int n )
{
	bool repeat = false;
	int i;
	double total = 0;
	for ( i = 0; i < length; i++ )
	{
		total += genes[ i ];
	}
	for ( i = 0; i < getLength(); i++ )
	{	
		genes[ i ] /= total;
	}
}

/* compute the problem-specific objective function
   In this problem, Sharpe Ratio is used as an objective function:
   SharpeRatio = ( ExpectedReturn - RiskFreeRate ) / StandardDeviation
   @return double 
*/
double Chromosome::objectiveFunction()
{
	double portfolioExpReturn = 0, variance = 0, myfitness = 0;

	for ( int i = 0; i < Parameter::genes; i++ )
	{
		portfolioExpReturn += Parameter::expReturn[ i ] * genes[ i ];
		variance += sqr( genes[ i ] ) * sqr( Parameter::stdDevReturn[ i ] );
		for ( int j = i + 1; j < Parameter::genes; j++ )
		{
			double covariance = Parameter::correlation[ i ][ j ] * Parameter::stdDevReturn[ i ] * Parameter::stdDevReturn[ j ];
			variance += 2 * genes[ i ] * genes[ j ] * covariance;
		}
	}
	myfitness = ( portfolioExpReturn - Parameter::riskfree ) / sqrt( variance );
	return myfitness;
}
